use crate::prelude::*;
use anyhow::Result;

use std::collections::HashMap;
use std::fmt::{Debug, Formatter};
use std::sync::{Arc, Weak};

pub enum UiElement {
    Button(Weak<Button>),
    Grid(Weak<Grid>),
    Label(Weak<Label>),
    MultiLineLabel(Weak<MultiLineLabel>),
    MultiLineTextField(Weak<MultiLineTextField>),
    TextField(Weak<TextField>),
    Icon(Weak<Icon>),
    ProgressBar(Weak<ProgressBar>),
    GuiSnippet(Weak<GuiSnippet>),
}

impl Debug for UiElement {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match self {
            UiElement::Button(_) => write!(f, "Button"),
            UiElement::Grid(_) => write!(f, "Grid"),
            UiElement::Label(_) => write!(f, "Label"),
            UiElement::MultiLineLabel(_) => write!(f, "MultiLineLabel"),
            UiElement::MultiLineTextField(_) => write!(f, "MultiLineTextField"),
            UiElement::TextField(_) => write!(f, "TextField"),
            UiElement::Icon(_) => write!(f, "Icon"),
            UiElement::ProgressBar(_) => write!(f, "ProgressBar"),
            UiElement::GuiSnippet(_) => write!(f, "GuiSnippet"),
        }
    }
}

macro_rules! impl_from {
    ($data_type:ident $(< $($lt:lifetime),+ >)? ) => {
        impl From<&Arc<$data_type$(<$($lt,)+>)?>> for UiElement {
            fn from(v: &Arc<$data_type$(<$($lt,)+>)?>) -> Self {
                UiElement::$data_type(Arc::downgrade(v))
            }
        }
    };
}

impl_from!(Button);
impl_from!(Grid);
impl_from!(Label);
impl_from!(MultiLineLabel);
impl_from!(MultiLineTextField);
impl_from!(TextField);
impl_from!(Icon);
impl_from!(ProgressBar);
impl_from!(GuiSnippet);

pub trait GetElement<T> {
    fn element(&self, id: &str) -> Result<Arc<T>>;
}

macro_rules! impl_element {
    ($data_type:ident $(< $($lt:lifetime),+ >)? ) => {
        impl GetElement<$data_type$(<$($lt,)+>)?> for HashMap<String, UiElement> {
            fn element(&self, id: &str) -> Result<Arc<$data_type$(<$($lt,)+>)?>> {
                match self.get(id) {
                    Some(element) => match element {
                        UiElement::$data_type(weak) => match weak.upgrade() {
                            Some(t) => Ok(t),
                            None => return Err(
                                anyhow::Error::msg(
                                    format!(
                                        "Invalid weak reference: {}",
                                        stringify!($data_type$(<$($lt,)+>)?)
                                    )
                                )
                            )
                        },
                        _ => panic!("wrong element type: {:?}", element),
                    },
                    None => panic!("no element with id ({}) found", id),
                }
            }
        }
    };
}

impl_element!(Button);
impl_element!(Grid);
impl_element!(Label);
impl_element!(MultiLineLabel);
impl_element!(MultiLineTextField);
impl_element!(TextField);
impl_element!(Icon);
impl_element!(ProgressBar);
impl_element!(GuiSnippet);
